"""
Provides:
  1. Library for testing factor strategies
  2. A lot of different statistical metrics for portfolios
  3. Fancy visualization of results

Source: https://github.com/atomtosov/pqr/tree/dev
"""

from .benchmarks import *
from .factor_models import *
from .factors import *
from .metrics import *
from .portfolios import *
from .preprocessing import *
from .tests import *
from .visualization import *
